#include "822lib.h"
#include "uotime.h"
#include "str2num.h"
#include <string.h>

#define DEBUGLEVEL 0
#if DEBUGLEVEL  > 0
#include <stdio.h>
#endif

int
r822_date(uo_sec70_t *out, uostr_t *in)
{
	#define MAXTOK 15
	uostr_t token=UOSTR_INIT;
	char *ar_type[MAXTOK];
	char *ar_data[MAXTOK];
	char *p,*e;
	int state=0;
	int tokens=0;
	int er;
	int i;
	int tz;
	uo_datetime_t dt;
	memset(&dt,0,sizeof(dt));
	tz=-1;
	er=r822_lex(&token,in);
	if (er) {
		uostr_freedata(&token);
		return er;
	}
	p=token.data;
	e=token.data+token.len;
	while (p!=e) {
		char *q;
		q=++p; /* skip \0 */
		while (*p && p!=e) p++;
		if ((p[-1]!='S' || q[0]!=' ') && p[-1]!='C') {
			if (tokens==MAXTOK)  {
				uostr_freedata(&token);
				return R822_ERR_PARSE;
			}
			ar_type[tokens]=p-1;
			ar_data[tokens]=q;
#if DEBUGLEVEL > 1
printf("token %d: %c '%*.*s'\n",tokens, ar_type[tokens][0],
	ar_type[tokens]-ar_data[tokens],ar_type[tokens]-ar_data[tokens],
	ar_data[tokens]);
#endif
			tokens++;
		}
	}
	/* states:
	 * 0   1  2 3   4    5 67 89  10
	 * [Sun,] 2 Jan 2000 11:08:24 -0000
	 * [Sun,] 2 Jan 2000 11:08:24 GMT [DST]
	 * state 0: got nothing
	 *       1: got day
	 *       2: got ,
	 *       3: got numeric day
	 * btw, unix uses:
	 *  Sun Jan  2 12:49:31 GMT 2000
	 */
	state=0;
	i=0;
	while (i!=tokens) {
      redo:
		if (state==0) {
			if (ar_type[i][0]!='W') goto parse_error;
			if (ar_data[i][0]<'0' || ar_data[i][0]>'9')
				state=1;
			else {
				state=2;
				goto redo;
			}
		} else if (state==1) {
			if (ar_type[i][0]!='S' ||ar_data[i][0]!=',') { state=2; goto redo; }
			state=2;
		} else if (state==2 || state==5 || state==7 || state==9) { 
			/* got comma, year, colon, second colon */
			unsigned long ul;
			int l;
			short *ptr=NULL; /* keep compiler quiet */
			unsigned short mx=-1,mn=9999;
			if (state==2) { mn=1; mx=31; ptr=&dt.day; }
			if (state==5) { mn=0; mx=23; ptr=&dt.hour; }
			if (state==7) { mn=0; mx=60; ptr=&dt.min; }
			if (state==9) { mn=0; mx=61; ptr=&dt.sec; }
			if (*ar_type[i]!='W') goto parse_error;
			*ar_type[i]=0;
			l=str2ulong (ar_data[i], &ul, 10);
			if (l==-1 || ar_data[i][l] || ul<mn || ul>mx) goto parse_error;
			*ptr=ul;
			state++;
		} else if (state==3) { /* got numerical day */
			if (*ar_type[i]!='W') goto parse_error;
			*ar_type[i]=0;
			if (0==strcasecmp(ar_data[i],"jan")) dt.mon=0;
			else if (0==strcasecmp(ar_data[i],"feb")) dt.mon=1;
			else if (0==strcasecmp(ar_data[i],"mar")) dt.mon=2;
			else if (0==strcasecmp(ar_data[i],"apr")) dt.mon=3;
			else if (0==strcasecmp(ar_data[i],"may")) dt.mon=4;
			else if (0==strcasecmp(ar_data[i],"jun")) dt.mon=5;
			else if (0==strcasecmp(ar_data[i],"jul")) dt.mon=6;
			else if (0==strcasecmp(ar_data[i],"aug")) dt.mon=7;
			else if (0==strcasecmp(ar_data[i],"sep")) dt.mon=8;
			else if (0==strcasecmp(ar_data[i],"oct")) dt.mon=9;
			else if (0==strcasecmp(ar_data[i],"nov")) dt.mon=10;
			else if (0==strcasecmp(ar_data[i],"dec")) dt.mon=11;
			else goto parse_error;
			state++;
		} else if (state==4) { /* got mon */
			unsigned long ul;
			int l;
			if (*ar_type[i]!='W') goto parse_error;
			*ar_type[i]=0;
			l=str2ulong (ar_data[i], &ul, 10);
			/* the 9999 check is done for RFC compatability */
			/* 0 is explicitly allow, for y2k systems ... */
			if (l==-1 || ar_data[i][l] || ul>9999) goto parse_error;
			if (ul<50) ul+=2000; 
			if (ul<999) ul+=1900; /* fails 2050 */
			dt.year=ul;
			state++;
		} else if (state==6||state==8) { /* got hour , got min */
			if (*ar_type[i]!='S' || ar_data[i][0]!=':') {
				/* shouldn't happen, but some software leaves out 
				 * this seconds part ...
				 */
				if (state==8) { state+=2; dt.sec=0; goto redo; }
				goto parse_error;
			}
			state++;
		} else if (state==10) {
			char *t;
			if (*ar_type[i]!='W') goto parse_error;
			*ar_type[i]=0;
			t=ar_data[i];
			if ((*t=='+' || *t=='-') && t[1]>='0' && t[1]<='9') {
				long lo;
				int vz=1;
				int l;
				l=str2long (ar_data[i], &lo, 10);
				/* the 9999 check is done for RFC compatability */
				if (l==-1 || ar_data[i][l] || lo<-9999 || lo>9999) 
					goto parse_error;
				if (lo<0) {vz=-1; lo=-lo; }
				tz=vz*((lo%100)+(lo/100)*60);
			} else if (!strcasecmp(t,"UT") || !strcasecmp(t,"UTC")
				|| !strcasecmp(t,"GMT")) tz=0;
			else if (!strcasecmp(t,"BST")) tz = 60;
			else if (!strcasecmp(t,"CDT")) tz = -300;
			else if (!strcasecmp(t,"CET")) tz = 60;
			else if (!strcasecmp(t,"CST")) tz = -360;
			else if (!strcasecmp(t,"CEST")) tz = -300;
			else if (!strcasecmp(t,"EDT")) tz = -240;
			else if (!strcasecmp(t,"EET")) tz = 120;
			else if (!strcasecmp(t,"EST")) tz = -300;
			else if (!strcasecmp(t,"HKT")) tz = 480;
			else if (!strcasecmp(t,"IST")) tz = 120;
			else if (!strcasecmp(t,"JST")) tz = 540;
			else if (!strcasecmp(t,"MDT")) tz = -360;
			else if (!strcasecmp(t,"MET")) tz = 60;
			else if (!strcasecmp(t,"METDST")) tz = 120;
			else if (!strcasecmp(t,"MEST")) tz = 120;
			else if (!strcasecmp(t,"MST")) tz = -420;
			else if (!strcasecmp(t,"PDT")) tz = -420;
			else if (!strcasecmp(t,"PST")) tz = -480;
			/* else error ignored: too much garbage on this world */
			state++;
		} else if (state==10) {
			if (*ar_type[i]!='W') goto parse_error;
			*ar_type[i]=0;
			if (strcasecmp(ar_data[i],"DST")) 
				goto parse_error;
			tz+=60;
		}
		i++;
	}
	*out=uo_dt2sec70(&dt);
	*out-=(tz*60);

	uostr_freedata(&token);
	return R822_OK;
  parse_error:
#if DEBUGLEVEL  > 0
printf("parse error at state %d, before token %i, '%*.*s'\n",
	state,i, ar_type[i]-ar_data[i],ar_type[i]-ar_data[i],
	ar_data[i]);
#endif
	uostr_freedata(&token);
	return R822_ERR_PARSE;
}
