/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1994, 1995 Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   domains.c: Domainverwaltung fr MausConnect
*/

#include "config.h"
#include "mausconsys.h"

#include <stdio.h>
#include <syslog.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "domains.h"
#include "uoio.h"

static box_t *b_anker;

static BOOLEAN
dom_read_itb(const char *fname, int fd)
{
	box_t *b=NULL;
	uoio_t io;

	uoio_assign_r(&io,fd,read,0);
	while (1) {
		char *buf;
		ssize_t len;

		len=uoio_getdelim_zc(&io,&buf,'\n');
		if (len==0) { /* EOF */
			uoio_destroy(&io);
			return TRUE;
		} 
		if (len==-1) {
			log (LOG_NOTICE,_("cannot read %s: %s"),fname,strerror(errno));
			uoio_destroy(&io);
			return FALSE;
		}
		if (buf[len-1]!='\n') {
			log (LOG_NOTICE,_("cannot read %s: last line incomplete"),fname);
			uoio_destroy(&io);
			return FALSE;
		}
		buf[--len]=0;

		switch(*buf)
		{
		case '*': 
			b=XMALLOC(sizeof(box_t));
			b->next=b_anker;
			b_anker=b;
			b->kuerzel=XSTRDUP(buf+1);
			b->vglkuerzel=entumlauten1(buf+1);
			XNULLCHECK(b->vglkuerzel);
			b->secdomain=NULL;
			b->primdomain=NULL;
			b->orga=NULL;
			break;
		case 'D':
			if (buf[1]=='.') {
				char *buf2;
				buf2=xconcat(b->kuerzel,buf+1,NULL);
				b->primdomain=entumlauten1(buf2);
				free(buf2);
			} else {
				b->primdomain=entumlauten1(buf+1);
			}
			XNULLCHECK(b->primdomain);
			break;
		case 'Z':
			{
				mc_list_t *ele=XMALLOC(sizeof(mc_list_t));
				ele->ext=NULL;
				ele->next=b->secdomain;
				b->secdomain=ele;
				if (buf[1]=='.') {
					char *buf2;
					buf2=xconcat(b->kuerzel,buf+1,NULL);
					ele->data=entumlauten1(buf2);
					free(buf2);
				} else {
					ele->data=entumlauten1(buf+1);
				}
				XNULLCHECK(ele->data);
			}
			break;
		case 'N':
			b->orga=XSTRDUP(buf+1);
			break;
		case 'T':
			if (b->orga) {
				char *t=xconcat(b->orga, " ", buf+1,NULL);
				free(b->orga);
				b->orga=t;
			}
		}
	}
	return TRUE;
}

static BOOLEAN
dom_read_domains(const char *fname, int fd)
{
	uoio_t io;
	uoio_assign_r(&io,fd,read,0);
	while (1) {
		char *buf;
		int ret;
		char *p, *q;
		box_t *b;
		mc_list_t *sec;

		ret=uoio_getdelim_zc(&io,&buf,'\n');
		if (ret==-1) {
			log (LOG_NOTICE,_("cannot read %s: %s"),fname,strerror(errno));
			uoio_destroy(&io);
			return FALSE;
		}
		if (ret==0) {
			uoio_destroy(&io);
			return TRUE;
		}
		if (buf[ret-1]!='\n') {
			log (LOG_NOTICE,_("cannot read %s: last line incomplete"),fname);
			uoio_destroy(&io);
			return FALSE;
		}
		buf[--ret]=0;

		p=strchr(buf,'\r');
		if (p)
			*p='\0';

		if (isspace(*buf)) {
			if (!b_anker)
				continue;
			p=buf;
			while(isspace(*p))
				p++;
			if (*p=='O')
				b_anker->orga=strdup(p+1);
			continue;
		}
		p=strpbrk(buf,"\t ");
		if (p)
			*p++='\0';
		b=XMALLOC(sizeof(box_t));
		b->next=b_anker;
		b_anker=b;
		b->kuerzel=XSTRDUP(buf);
		b->vglkuerzel=entumlauten1(buf);
		XNULLCHECK(b->vglkuerzel);
		b->secdomain=NULL;
		b->primdomain=NULL;
		b->orga=NULL;
		if (!p)
			continue;
		while (isspace(*p))
			p++;
		q=strpbrk(p,"\t ");
		if (q)
			*q++='\0';
		b->primdomain=XSTRDUP(p);

		p=q;
		while (p)
		{
			while (isspace(*p))
				p++;
			q=strpbrk(p,"\t ");
			if (q)
				*q++='\0';
			sec=XMALLOC(sizeof(mc_list_t));
			sec->ext=NULL;
			sec->next=b->secdomain;
			b->secdomain=sec;
			sec->data=XSTRDUP(p);
			p=q;
		}
	}
}

BOOLEAN 
dom_init(void)
{
	uostr_t fname=UOSTR_INIT;
	int fd;

	uostr_xdup_cstr(&fname,opts.spoolbase);	
	uostr_xadd_char(&fname,'/');
	uostr_xadd_cstr(&fname,action);
	uostr_xadd_cstr(&fname,"/infofiles/ITS");
	uostr_x0(&fname);
	fd=open(fname.data,O_RDONLY);
	if (fd!=-1) {
		if (!dom_read_itb(fname.data,fd)) {
			close(fd);
			uostr_freedata(&fname);
			return FALSE;
		}
		uostr_freedata(&fname);
		close(fd);
	} else {
		uostr_xdup_cstr(&fname,opts.spoolbase);	
		uostr_xadd_char(&fname,'/');
		uostr_xadd_cstr(&fname,action);
		uostr_xadd_cstr(&fname,"/etc/domains");
		uostr_x0(&fname);
		fd=open(fname.data,O_RDONLY);
		if (fd==-1) {
			uostr_t fname2=UOSTR_INIT;
			int s=errno;
			uostr_xdup_cstr(&fname2,opts.spoolbase);	
			uostr_xadd_char(&fname2,'/');
			uostr_xadd_cstr(&fname2,action);
			uostr_xadd_cstr(&fname2,"/domains");
			uostr_x0(&fname2);
			fd=open(fname2.data,O_RDONLY);
			uostr_freedata(&fname2);
			if (fd==-1) {
				log (LOG_NOTICE,_("cannot read %s: %s"),fname.data,strerror(s));
				uostr_freedata(&fname);
				return FALSE;
			}
		}
		if (!dom_read_domains(fname.data,fd)) {
			uostr_freedata(&fname);
			return FALSE;
		}
	}

	return TRUE;
}

box_t *
dom_find_boxkuerzel(const char *kuerzel)
{
	char buf[25];
	box_t *akt;
	char *p;
	/* Mauskrzel sind hchstens 8 Zeichen, und enthalten hoffentlich
	 * niemals mehr als 8 Umlaute :-). Aber wir sind vorsichtig und
	 * erlauben 12 Zeichen (Umlautform) und 24 (entumlautet).
	 */
	if (strlen(kuerzel)>12)
		return NULL;
	entumlauten(kuerzel,buf);
	p=buf+strlen(buf)-1;
	/* trailing spaces entsorgen */
	while (p>=buf) {
		if (*p=='\t' || *p==' ')
			*p--=0;
		else
			break;
	}
	for (akt=b_anker;akt;akt=akt->next)
	{
		if (!strcasecmp(buf,akt->vglkuerzel))
			return akt;
	}
	return NULL;
}

box_t *
dom_find_domain(const char *domain)
{
	box_t *akt;
	mc_list_t *sec;
	char *canon=entumlauten1(domain);
	XNULLCHECK(canon);
	str_lwr(canon);
	
	for (akt=b_anker;akt;akt=akt->next)
	{
		if (akt->primdomain 
			&& *canon==tolower((unsigned char) *akt->primdomain) 
			&& !strhalfcasecmp(akt->primdomain,canon))
		{
			free(canon);
			return akt;
		}
		for (sec=akt->secdomain;sec;sec=sec->next)
		{
			if (*canon==tolower((unsigned char) *sec->data)
				&& !strhalfcasecmp(sec->data,canon))
			{
				free(canon);
				return akt;
			}
		}
	}
	free(canon);
	return NULL;
}
