
/* 
 * He who invented this should be shot on the spot.
 * "not invented here".
 * What's so wrong about using standard unicode tables and the mappings?
 */

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/fcntl.h>
#include "uostr.h"
#include "uoio.h"
#include "itzlib.h"
#include "str2num.h"


#define NOCHAR ((unsigned int)0xffff)
	/* values below 32 are for internal use only, see itzlib.h */
#define FIRSTCHARSET 32
#define LASTCHARSET 126
#define MAXCHARS 65536
#define BLOCKSIZE 1024
#define NUMBLOCKS 64
struct charset {
	char *name;
	char *kuerzel;
	unsigned int cs[256];
};
struct charset *charsets;
struct itz_char {
	unsigned int num;
	char *sgml;
	char *ersatz;
};
static struct itz_char *blocks[NUMBLOCKS];	/*64 blocks of 1024 (BLOCKSIZE) */

static unsigned int *overhead;	/* indirect pointer */

const char *itz_get_ersatz(unsigned int c) {
	int o,block,ind; 
	o=overhead[c]; 
	block=o/BLOCKSIZE; 
	ind=o%BLOCKSIZE;
	return blocks[block][ind].ersatz;
}
const char *itz_get_sgml(unsigned int c) {
	int o,block,ind; o=overhead[c]; block=o/BLOCKSIZE; ind=o%BLOCKSIZE;
	return blocks[block][ind].sgml;
}
static void impossible (const char *s) { write (2, s, strlen (s)); write (2, "\n", 1); _exit (1); }

int
itz_walk_charset(int i, const char **s, const char **t)
{
	i++;
	if (i<0) return -1;
	if (i>LASTCHARSET) return -1;
	while (!charsets[i].name) {
		i++;
		if (i>LASTCHARSET) return -1;
	}
	*s=charsets[i].kuerzel;
	*t=charsets[i].name;
	return i;
}

static void
add_char_to_charsets (unsigned int num, char *line)
{
	unsigned long u;
	if (2 != str2ulong (line, &u, 16) || line[2] != '=')
		impossible ("malformed C-line");
	line += 3;
	while (*line) {
		int c = *line;
		if (c < FIRSTCHARSET || c > LASTCHARSET)
			impossible ("illegal charset number in C-line");
		/* take first entry if we have more than one */
		/* and redefining 0 is not a good idea */
		if (u && NOCHAR==charsets[c].cs[u]) 
			charsets[c].cs[u] = num;
		line++;
	}
}

static void
work_one (unsigned int num, char *sgml, char *ersatz)
{
	static int block = 0;
	static int bindex = 0;
	struct itz_char *i;
	if (bindex == BLOCKSIZE) {
		bindex = 0;
		block++;
	}
	if (block == NUMBLOCKS)
		impossible ("too many blocks in ITZ");
	if (!blocks[block])
		blocks[block] = calloc (sizeof (struct itz_char), BLOCKSIZE);
	if (!blocks[block])
		impossible ("out of memory");
	i = blocks[block];
	i += bindex;
	i->num = num;
	i->ersatz = ersatz;
	i->sgml = sgml;
	overhead[num] = block * BLOCKSIZE + bindex;
	bindex++;
}

int
itz_parse (const char *fname)
{
	int fd;
	uoio_t r;
	uoio_t e;
	int en;
	ssize_t l;
	char *s;
	int t_charsetnum = -1;
	char *t_charsetname = 0;
	char *t_charsetkuerzel = 0;
	unsigned int t_charnum = NOCHAR;
	char *t_ersatz = 0;
	char *t_sgml = 0;
	if (charsets) impossible("ITZ initialized twice");
	charsets=calloc(sizeof(struct charset),LASTCHARSET+1);
	overhead=calloc(sizeof(unsigned int),MAXCHARS);
	if (!overhead) goto oom;
	if (!charsets) goto oom;

	/* setup internal special character sets */
	{ 
		union { char *ws; const char *cs; } u; /* disqualify*/
		u.cs="ucs4";
		charsets[0].kuerzel=u.ws;
		u.cs="ISO 10646 Universal Character Set (UCS)";
		charsets[0].name=u.ws;
		u.cs="ucs2";
		charsets[1].kuerzel=u.ws;
		u.cs="Unicode";
		charsets[1].name=u.ws;
	}

	if (!fname) return -1;
	fd = open (fname, O_RDONLY);
	if (fd == -1)
		return -1;
	uoio_assign_r (&r, fd, read, 0);
	uoio_assign_w (&e, 2, write, 0);
	/* get list of charsets */
	while (1) {
		l = uoio_getdelim_zc (&r, &s, '\n');
		if (-1 == l) {
			goto ioerror;
		}
		if (0 == l || s[l - 1] != '\n') {
			uoio_write_cstr (&e, "ITZ unvollstndig\n");
			uoio_flush (&e);
			goto cleanup;
		}
		s[--l] = 0;
		if (!l)
			continue;			/* silently ignore illegal empty line */
		if (s[l - 1] == '\r')
			s[--l] = 0;
		switch (*s) {
		case 'Z':				/* "Id fr diesen Zeichensatz" */
			if (t_charsetnum!=-1) {
				int i;
				(charsets[t_charsetnum]).name = t_charsetname;
				charsets[t_charsetnum].kuerzel = t_charsetkuerzel;
				for (i=0;i<256;i++) charsets[t_charsetnum].cs[i]=NOCHAR;
			}
			t_charsetnum = -1;
			t_charsetname = 0;
			t_charsetkuerzel = malloc (l);
			if (!t_charsetkuerzel)
				goto oom;
			memcpy (t_charsetkuerzel, s + 1, l);
			continue;
		case 'I':{				/* "Krzel fr diesen Zeichensatz" */
				char x;
				x=s[1];
				if (x<FIRSTCHARSET || x>LASTCHARSET || s[2]) {
					uoio_write_cstr (&e, "illegal charset_id in ITZ: ");
					uoio_write_cstr (&e, s+1);
					uoio_write_cstr (&e, "\n");
					uoio_flush (&e);
					_exit(1);
				}
				t_charsetnum = x;
			}
			continue;
		case 'B':				/* Bitweite: b7 oder b8 */
			continue;
		case 'D':				/* Langer Name des Zeichensatzes */
			t_charsetname = malloc (l);
			if (!t_charsetname)
				goto oom;
			memcpy (t_charsetname, s + 1, l);
			continue;
		case 'V':
			continue;
		}
		break;
	}
	if (t_charsetnum) {
		int i;
		charsets[t_charsetnum].name = t_charsetname;
		charsets[t_charsetnum].kuerzel = t_charsetkuerzel;
		for (i=0;i<256;i++) charsets[t_charsetnum].cs[i]=NOCHAR;
	}
	t_charsetnum = 0;
	t_charsetkuerzel = 0;
	t_charsetname = 0;

	while (1) {
		unsigned long u;
		/* process last line, has been read above or below */
		switch (*s) {
		case '#':				/* "Beginn der Definition eines Unicode-Zeichens" */
			/* "codeposition des Zeichens in 4 Hex-Ziffern" */
			if (t_charnum != NOCHAR)
				work_one (t_charnum, t_sgml, t_ersatz);
			else {
				free (t_sgml);
				free (t_ersatz);
			}
			t_sgml = 0;
			t_ersatz = 0;
			if (0 == str2ulong (s + 1, &u, 16))
				goto illegal_charid;
			if (u>0xffff) goto illegal_charid; /* sorry, only 16 bit plane */
			t_charnum = u;
		case 'N':				/* offizieller Unicode-Name des Zeichens */
			break;
		case 'S':				/* SGML-Name */
			t_sgml = malloc (l);
			if (!t_sgml)
				goto oom;
			memcpy (t_sgml, s + 1, l);
			break;
		case 'E':				/* Ersatzstring */
			t_ersatz = malloc (l);
			if (!t_ersatz)
				goto oom;
			memcpy (t_ersatz, s + 1, l);
			break;
		case 'C':{
				add_char_to_charsets (t_charnum, s + 1);
				break;
			}
		}

		/* get new line */
		l = uoio_getdelim_zc (&r, &s, '\n');
		if (-1 == l) {
			goto ioerror;
		}
		if (0 == l) break;
		if (s[l - 1] != '\n') {
			uoio_write_cstr (&e, "ITZ unvollstndig\n");
			uoio_flush (&e);
			goto cleanup;
		}
		s[--l] = 0;
		if (!l)
			continue;			/* silently ignore illegal empty line */
		if (s[l - 1] == '\r')
			s[--l] = 0;
	}
	uoio_destroy(&r);
	uoio_destroy(&e);
	close(fd);
	return 0;

  oom: errno=ENOMEM; return -1;
  illegal_charid:
	uoio_write_cstr (&e, "illegal char_id in ITZ: ");
	uoio_write_cstr (&e, s);
	uoio_write_cstr (&e, "\n");
	uoio_flush (&e);
	_exit(1);
  ioerror:
	en = errno;
	uoio_write_cstr (&e, "I/O-Error on ITZ: ");
	uoio_write_cstr (&e, strerror (en));
	uoio_write_cstr (&e, "\n");
	uoio_flush (&e);
	_exit (1);
  cleanup:
	uoio_destroy (&e);
	uoio_destroy (&r);
	close (fd);
	return -1;
}

struct conversion {
	short from;
	short to;
	unsigned int num[256]; /* NOCHAR: nixda */
	const char *ersatz[256];
	const char *sgml[256];
	struct conversion *next;
};
	/* used conversion is the first of the list */
static struct conversion *anker;
static unsigned int *ucs4hack;

short
itz_get_charset_by_kuerzel(const char *s)
{
	int i;
	for (i=0;i<=LASTCHARSET;i++) {
		if (charsets[i].kuerzel && 0==strcasecmp(s,charsets[i].kuerzel))
			return i;
	}
	return -1;
}

/* performance is no problem in this case, it's lost anyway */
static uostr_t * 
itz_convert_ucs4(uostr_t *to, uostr_t *from, int width)
{
	char *p;
	unsigned int *num=anker->num;
	char *e=from->data+from->len;
	int lim;

	/* pathological case ... */
	if (anker->from==anker->to) return uostr_dup_uostr(to,from);
	if (anker->from==CHARSET_UCS4 && anker->to==CHARSET_UCS2) {
		/* easy ... */
		char *data;
		if (!uostr_need(to,from->len/2)) return 0;
		data=to->data;
		p=(from->data);
		while (e-p>3) {
			p+=2;
			*data++=*p++;
			*data++=*p++;
		}
		if (e!=p) {
			errno=EINVAL;
			uostr_freedata(to);
			return 0;
		}
  		to->len=data-to->data;
		return to;
	}
	if (anker->from==CHARSET_UCS2 && anker->to==CHARSET_UCS4) {
		/* easy ... */
		char *data;
		if (!uostr_need(to,from->len*2)) return 0;
		data=to->data;
		p=(from->data);
		while (e-p>1) {
			*data++=0;
			*data++=0;
			*data++=*p++;
			*data++=*p++;
		}
		if (e!=p) {
			errno=EINVAL;
			uostr_freedata(to);
			return 0;
		}
  		to->len=data-to->data;
		return to;
	}
	lim=width-1;

	if (anker->from==CHARSET_UCS4 || anker->from==CHARSET_UCS2) {
		char *data;
		/* advise uostr library that we need at least that much memory */
		if (!uostr_need(to,from->len/4)) return 0;
		data=to->data;
		p=(from->data);
		while (e-p>lim) {
			unsigned long x;
			if (width==4) {
				x=((unsigned int)(unsigned char) (*p++))<<24;
				x+=((unsigned int)(unsigned char) (*p++))<<16;
			} else
				x=0;
			x+=((unsigned int)(unsigned char) (*p++))<<8;
			x|=((unsigned int)(unsigned char) (*p++));
			if (x>65536) {
				errno=EINVAL;
				uostr_freedata(to);
				return 0;
			}
			if (ucs4hack[x]!=NOCHAR) {
				*data++=ucs4hack[x];
			} else {
				/* ersatztext */
				int need_sgml;
				const char *ers;
				size_t l;
				need_sgml=0;
				ers=itz_get_ersatz(x);
				if (!ers) {
					ers=itz_get_sgml(x);
					if (ers) need_sgml=1;
					else
						ers="?"; /* last chance */
				}
				l=strlen(ers);
				/* we need some more space. tell library to give us that */
				to->len=data-to->data;
				if (!uostr_needmore(to,l+need_sgml*2)) return 0;
				data=to->data+to->len;
				if (need_sgml) *data++='&';
				switch(l) {
				default: memcpy(data,ers,l); data+=l; break;
					/* critical point seems to be 6 */
				case 6: *data++=*ers++;
				case 5: *data++=*ers++;
				case 4: *data++=*ers++;
				case 3: *data++=*ers++;
				case 2:	*data++=*ers++;
				case 1: *data++=*ers++;
					break;
				}
				if (need_sgml) *data++=';';
				*data='Z';
			}
		}
		if (e!=p) {
			errno=EINVAL;
			uostr_freedata(to);
			return 0;
		}
  		to->len=data-to->data;
	} else {
		char *data;
		/* advise uostr library that we need at least that much memory */
		if (!uostr_need(to,from->len*4)) return 0;
		uostr_cut(to,0);
		data=to->data;
		p=(from->data);
		while (e!=p) {
			unsigned int c=num[(int)(unsigned char)(*p++)];
			if (width==4) {
				*data++=0; /* remember: we only can handle 16 bit */
				*data++=0; /* remember: we only can handle 16 bit */
			}
			*data++=(c>>8);
			*data++=(c&0xffff);
		}
  		to->len=data-to->data;
	}
	return to;
}

#undef NO_ERSATZ /* 15% more performance */
uostr_t * 
itz_convert_uostr(uostr_t *to, uostr_t *from)
{
	char *p;
	unsigned int *num=anker->num;
	char *e=from->data+from->len;
	char *data;


	/* special cases for something to !8bit conversions */
	if (anker->to==CHARSET_UCS4) return itz_convert_ucs4(to,from,4);
	if (anker->from==CHARSET_UCS4) return itz_convert_ucs4(to,from,4);
	if (anker->to==CHARSET_UCS2) return itz_convert_ucs4(to,from,2);
	if (anker->from==CHARSET_UCS2) return itz_convert_ucs4(to,from,2);

	/* advise uostr library that we need at least that much memory */
	if (!uostr_need(to,from->len)) return 0;
	uostr_cut(to,0);

	data=to->data;
	p=(from->data);

	while (1) {
		int translated;
		int c;
		/* this very ugly piece of code is the fastest i came up with so far.
		 * with unrolled loop and duffs device it's about 20 percent faster
		 * than before. 
		 * Typically charset conversion will be the slowest part of the 
		 * program, so optimizing it is simply a "must". Note that if you
		 * define NO_ERSATZ the code will be 15% faster: Thats the price for
		 * a comparision and a jump. It would be even faster if it did the
		 * replacement "in place", like the quark BBS uses to do it: 
		 * processors seem to dislike featuritis and _punish_.
		 * But note that in-place-replacement is not an option if using
		 * ersatz/replacement strings or converting to utf8.
		 */
#ifdef NO_ERSATZ
#define WORK do {c=(unsigned char)*p++; *data++=num[c];}while(0)
#else
#define WORK do {c=(unsigned char)*p++; if ((translated=num[c])==NOCHAR) goto out; *data++=translated;}while(0)
#endif
		switch ((e-p)%6) {
		while (1) {
		default: 
				if (e==p) goto do_return;
				WORK;
		case 5: WORK;
		case 4: WORK;
		case 3: WORK;
		case 2: WORK;
		case 1: WORK;
		}
		} /* switch */
#ifndef NO_ERSATZ
	  out:
		/* c is the char we have to translate in a more complicated way */
		{
			const char **ersatztab=anker->ersatz;
			const char **sgmltab=anker->sgml;
			int need_sgml;
			const char *ers;
			size_t l;
			need_sgml=0;
			ers=ersatztab[c];
			if (!ers) {
				need_sgml=1;
				ers=sgmltab[c];
			}
			l=strlen(ers);
			/* we need some more space. tell library to give us that */
  			to->len=data-to->data;
			if (!uostr_needmore(to,l+need_sgml*2)) return 0;
			data=to->data+to->len;
			if (need_sgml) *data++='&';
			switch(l) {
			default: memcpy(data,ers,l); data+=l; break;
				/* critical point seems to be 6 */
			case 6: *data++=*ers++;
			case 5: *data++=*ers++;
			case 4: *data++=*ers++;
			case 3: *data++=*ers++;
			case 2:	*data++=*ers++;
			case 1: *data++=*ers++;
				break;
			}
			if (need_sgml) *data++=';';
			*data='Z';
		}
#endif
	}
  do_return:
  	to->len=data-to->data;
	return to;
}

int
itz_setup_conversion(const char *sfrom, const char *sto)
{
	struct conversion *neu;
	short from,to;
	int tmp[256];
	/* don't want to cache that, see ucs4hack */
	if (anker && anker->from==CHARSET_UCS4) {
		neu=anker;
		anker=anker->next;
		free(neu);
	}
	from=itz_get_charset_by_kuerzel(sfrom);
	if (-1==from) {errno=ENOENT;return -1;}

	to=itz_get_charset_by_kuerzel(sto);
	if (-1==to) {errno=ENOENT;return -1;}

	/* find something in cache */
	if (from!=CHARSET_UCS4) {
		struct conversion *dies,*last; /* hate C++ for this! */
		for (dies=anker,last=0;dies;last=dies,dies=dies->next) {
			if (dies->to==to && dies->from==from) {
			/* reorder */
			if (last) last->next=dies->next;
				else anker=dies->next;
				dies->next=anker;
				anker=dies;
				return 0;
			}
		}
	}

	neu=calloc(1,sizeof(struct conversion));
	if (!neu) {errno=ENOMEM;return -1;}
	memset(tmp,0,sizeof(tmp));
	neu->from=from;
	neu->to=to;
	neu->next=anker;
	anker=neu;

	if (from==CHARSET_UCS4 || from==CHARSET_UCS2) {
		if (to!=CHARSET_UCS4 && to!=CHARSET_UCS2) {
			unsigned int i;
			ucs4hack=malloc(65536*sizeof(int));
			if (!ucs4hack) { errno=ENOMEM; return 0; }
			for (i=0;i<65536;i++) ucs4hack[i]=NOCHAR;
			for (i=0;i<256;i++) {
				unsigned int uc=charsets[to].cs[i];
				/* don't overwrite 0x0a with 0x8a. */
				if (ucs4hack[uc]==NOCHAR 
					|| ucs4hack[uc]>i)
					ucs4hack[uc]=i;
			}
		}
	} else if (to==CHARSET_UCS4||to==CHARSET_UCS2) {
		unsigned int i;
		for (i=0;i<256;i++)
			neu->num[i]=charsets[from].cs[i];
	} else {
		unsigned int i;
		int pos_beta=0;
		int pos_sz=0;
		for (i=0;i<256;i++) {
			tmp[i]=charsets[from].cs[i]; 
			neu->num[i]=NOCHAR; /* nix */
			neu->ersatz[i]=0;
			neu->sgml[i]=0;
		}
		for (i=0;i<256;i++) {
			int j;
			int uc=charsets[to].cs[i];
			for (j=0;j<256;j++) {
				if (tmp[j]==uc) {
					/* -8859-x 0 and 80 are the same in ITZ??? */
					if (neu->num[j]==NOCHAR) {
						if (tmp[j]==0x03B2) pos_beta=j;
						if (tmp[j]==0x00DF) pos_sz=j;
						neu->num[j]=i;
					}
					break;
				}
			}
		}
		neu->num[0]=0; /* everything else leads to madness */
		for (i=0;i<256;i++) {
			if (neu->num[i]==NOCHAR) {
#ifdef NO_ERSATZ
				neu->num[i]=i;
#else
				int x;
				x=charsets[from].cs[i];
				if (x==NOCHAR) neu->num[i]=i; /* no way */
				else {
					const char *e,*s;
					e=itz_get_ersatz(charsets[from].cs[i]);
					if (e && 0==strcmp(e,"beta")) { /* fix braindamage */
						int j;
						e=0;
						for (j=0;j<256;j++) {
							if (charsets[from].cs[j]==0x00DF) { /* sz */
								/* charset without beta sign, but sz, fine. */
								static char hack[2]={0xdf,0};
								e=hack;
								break;
							}
						}
					}
					if (e)
						neu->ersatz[i]=e;
					s=itz_get_sgml(charsets[from].cs[i]);
					if (s) 
						neu->sgml[i]=s;
					if (!s && !e)
						neu->num[i]=i; /* try something */
				}
#endif
			}
		}
	}
	return 0;
}

