/* metamsg.h: metamsg_t definition and prototypes */
/*
    Copyright (C) 1994 1995 1996 1997 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@ohse.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
*/
#ifndef METAMSG_H
#define METAMSG_H

#include "uotime.h"

typedef struct metaliststruct
{
	char *text; /* data */
	char *ext;  /* extension */
	struct metaliststruct *next;
} metalist_t;

typedef struct metamsg
{
	/* Ist es Mail oder News? */
	int is_mail;
	/* Envelope der Nachricht, nur bei MAIL */
	char *     env_from;       /* physikalischer Absender */
	char *     env_to;         /* Zieladresse dieser Nachricht */
	/* Headerangaben */
	metalist_t *head_to;         /* Empfaenger */
	metalist_t *head_from;       /* Absenderadresse + -name */
	metalist_t *reply_to;        /* Empfnger einer Antwort (adresse+name) */
	metalist_t *sender;          /* tatschlicher Absender (adresse+name) */
	metalist_t *cc;              /* CC-Liste, fr rfc2maus --read-header */
	metalist_t *bcc;             /* BCC-Liste, fr rfc2maus --read-header */

	metalist_t *newsgroups;      /* Liste der Newsgroups */
	metalist_t *followup_to;     /* Liste der Followup-Groups */

	/* Message-ID */
	char *message_id;
	char *maus_id;               /* ID im MausNet */
	char *maus_special_id;       /* ITG, LOG, HEAD, ... */
	char *transfer_id;           /* Uebergangs-ID zwischen User und MAUS */
	metalist_t *references;      /* Usenetreihenfolge! */

	/* Sonstige wichtige Headerzeilen */
	uo_sec70_t date; /* -1: ungueltig */
	char *distribution;
	char *subject;

	char *maus_status;

	char *also_control;
	char *approved;
	char *content_type;
	char *content_transfer_encoding;
	char *content_description;
	char *content_id;
	char *control;
	char *mime_version;
	char *received;
	struct tm *tm_expires;
	char *keywords;
	char *organization;
	char *path;
	char *summary;
	metalist_t *supersedes;
	metalist_t *replaces;
	char *comments;
	char *x_gateway; /* MausConnect blah fasel */
	char *x_format;  /* Usprungsformat der Nachricht */
	char *x_error;   /* Fehlermeldungen */
	char *x_mailer;  /* Angabe der Software des Ursprungssystems */
	metalist_t *return_receipt_to;
	char *x_to;      /* FIDO-Style To in News */
	char *quark_a_zeile; /* lokaler Empfnger-Benutzername der Quark (auch andere Boxen haben so einen */

	/* von mauscon erzeugte Debug-Header. */
	metalist_t *trace_header;

	/* Still unused */
	metalist_t *unknown_headers;
	metalist_t *header_garbage;


	/* Pointer auf Pufferbereich mit dem Messagetext darin */
	/* Der Speicherbereich wird von meta_clean_message nicht angerhrt */
	uostr_t body;

	/* Zaehler, die von der Routine zum Anhaengen an den Body verwaltet
	 * werden.
	 */
	struct {
		unsigned long lines;
		unsigned long bytes;
		unsigned long achtbit; /* Zahl der 8bit-Zeichen */
	} bodystat;

	/* Mausflags (X-Zeile) */
	struct {
		enum {XR_NOTSET, XR_PLUS, XR_MINUS} exportflag;
	} mausflags;

	BOOLEAN mime_code_headers; /* Header mit q-p verschlsseln? */
	BOOLEAN mime_code_body; /* Body? */

	/* file wird *immer* in Ruhe gelassen - es ist Sache der
	 * hhergelegenen Routinen, die Datei zu schlieen.
	 * In any case message-id, envelope and/or newsgroups must be set!
	 */
	FILE *file; /* if not NULL read header and message from there */

} metamsg_t;

void 
meta_clean_message(metamsg_t *msg);


BOOLEAN 
meta_add_list(metalist_t **li, const char *text, const char *ext,int vorne);
BOOLEAN 
meta_add_trace(metamsg_t *msg, int grade, const char *text, ...);

	/* Ein metalist_t * als string schreiben. */
int rfc_metalist2adrheader(BOOLEAN mime, uostr_t *buf,metalist_t *list);

	/* ein Element mit Kopien der beiden Strings anlegen */
metalist_t *meta_create_element(const char *text, const char *ext);

	/* Einen String an den Body anfgen. Die Zeilen muessen mit \r\n 
	 * terminiert sein.
	 */
void meta_add_to_body(metamsg_t *msg, const char *s);
#define METAGMSG_BODY_GROW (64*1024)
static __inline__ void
meta_fast_add_to_body_crlf(metamsg_t *msg, const char *s, size_t llen)
{
	size_t used;

	msg->bodystat.bytes+=llen+2;

	used=msg->body.len;
	if (llen+1+2+used>msg->body.size)
	{
		size_t grow=llen+1+used-msg->body.size;
		if (grow < METAGMSG_BODY_GROW)
			grow=METAGMSG_BODY_GROW;
		else
			grow = ((grow + METAGMSG_BODY_GROW -1) /METAGMSG_BODY_GROW ) *METAGMSG_BODY_GROW;
		uostr_xneedmore(&msg->body,grow);
	}
	memcpy(msg->body.data+msg->body.len,s,llen+1);
	msg->body.len+=llen;
	msg->body.data[msg->body.len++]='\r';
	msg->body.data[msg->body.len++]='\n';
	msg->body.data[msg->body.len]='0';
}
#define GROW (64*1024)
static __inline__ void 
meta_add_to_body_crlf(metamsg_t *msg, const char *s, size_t llen)
{
	size_t used;
	char *q;

	used=msg->body.len;
	if (llen+2+1+used>msg->body.size)
	{
		size_t grow=llen+2+1+used-msg->body.size;
		if (grow < GROW)
			grow=GROW;
		else
			grow = ((grow + GROW -1) /GROW ) *GROW;
		uostr_xneedmore(&msg->body,grow);
	}
	q=msg->body.data+msg->body.len;
	if (llen) {
		switch (llen%8) {
			while (*s) {
		case 0: if (*s & 0x80) msg->bodystat.achtbit++; *q++=*s++;
		case 7: if (*s & 0x80) msg->bodystat.achtbit++; *q++=*s++;
		case 6: if (*s & 0x80) msg->bodystat.achtbit++; *q++=*s++;
		case 5: if (*s & 0x80) msg->bodystat.achtbit++; *q++=*s++;
		case 4: if (*s & 0x80) msg->bodystat.achtbit++; *q++=*s++;
		case 3: if (*s & 0x80) msg->bodystat.achtbit++; *q++=*s++;
		case 2: if (*s & 0x80) msg->bodystat.achtbit++; *q++=*s++;
		case 1: if (*s & 0x80) msg->bodystat.achtbit++; *q++=*s++;
			}
		}
	}
	*q++='\r';
	*q++='\n';
	*q='\0';
	msg->body.len=q-msg->body.data;
	msg->bodystat.lines++;
	msg->bodystat.bytes+=llen+2;
}
#undef GROW


	/* Eine Liste entfernen */
void meta_free_list (metalist_t **l);

void meta_free(metamsg_t *);
metamsg_t *meta_alloc(void);

#endif


