
/* Mauscon: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1992 - 1998 1999 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   mroute.c: Spoolt News+Mail Richtung MausNet
*/

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include "uogetopt.h"
#include <limits.h>
#include <syslog.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "errormail.h"
#include "rfc.h"
#include "route.h"
#include "setvars.h"

#ifndef STDIN_FILENO
#define STDIN_FILENO 0
#endif

static char *gm_addr;
static int is_routing=0;
static int got_cancels=0;

/* two dummies, linking srfcmail/srfcnews needs them */
int dont_use_nntp;
int dont_use_smtp;

struct empfaenger {
	char *adr;
	struct empfaenger *next;
};

static int opt_test=0;

static long get_transfer_id (const char *errors_to);
static int handle_mail(const char *inputfilename, const  char *from, const char *to);
static int handle_news(const char *inputfilename);
static int handle_rnews(const char *inputfilename);
static int handle_rsmtp(const char *inputfilename);
static int handle_rmail(const char *inputfilename, char **adressen);
static void	test_routing(void);
static void finish_rsmtp(char *from, struct empfaenger *emp, uostr_t *message);
static void cleanup_empfaenger(struct empfaenger **anker);


uostr_t mbuf;

static int
work_cancel(char *s, char *from)
{
	metalist_t *mlp,*ml=NULL;
	int troubles;
	if (!rfc_ids_to_metalist(&ml,s,&troubles)||troubles) {
		meta_free_list(&ml);
		log(LOG_WARNING,"can't convert %s",s);
		return 1;
	}
	mlp=ml;
	while (mlp) {
		got_cancels++;
		cancel_add(mlp->text,from);
		mlp=mlp->next;
	}
	meta_free_list(&ml);
	return 0;
}  /* work cancel */

static long
get_transfer_id (const char *error_to)
{
	int count;
	int err;
	long mid;
	static uostr_t midname=UOSTR_INIT;
	static uostr_t lockname=UOSTR_INIT;
	char str[255];
	FILE *f;

	uostr_xdup_cstrmulti(&midname,
		opts.spoolbase, "/", action, "/transfer.id",0);
	uostr_xdup_uostr(&lockname,&midname);
	uostr_x0(&midname); /* \0 */
	uostr_xadd_mem(&lockname,".LCK",5); /* \0 */
	
	count = 0;
	err = 0;
	while (1)
	{
		if (!lock_create(lockname.data,NULL))
			break;
		count++;
		if (opts.lock_tries && count>=opts.lock_tries)
		{
			errormail (NULL,
				error_to ?  "error spooling message"
				:	"error spooling message - sender unknown",
		   		"couldn't get a transfer id (spooling message failed)",
		   gm_addr, error_to ? (error_to) : (NULL) , NULL);
			exit (1);
		}
		if (opts.lock_sleep > 0)
			sleep (opts.lock_sleep);
	}
	f = fopen (midname.data, "r+");
	if (!f)
	{
		/* einfach waere es, jetzt 0 zurueckliefern. Haken dabei:
		   Falls die Datei mal irgendwie "verschwindet", wuerden neue
		   Nachrichten unter den alten IDs erzeugt, und zumindest der
		   Dupecheck der Quark II findet das nicht so toll.
		   Also nehmen wir einen von der aktuellen Uhrzeit bzw. Datum
		   abhaengigen Wert. Problem: User-IDs in MAUS duerfen nur
		   bis zu 10 Zeichen lang sein. Da der Prefix "mail" vor
		   Mails aber schon vier Zeichen belegt, kann der 32-Bit-Wert
		   von time() nicht direkt verwendet werden. Nimmt man aber den
		   Rest einer Division des time()-Rueckgabewerts durch eine
		   Million, sollten Kollisionen hinreichend unwahrscheinlich
		   sein.
		*/
		mid=time(NULL) % 1000000;
		f=fopen(midname.data,"w");
		if (!f)
		{
			errormail (NULL,
				error_to ?
			   		"error spooling message"
				:	"error spooling message - sender unknown",
			   "couldn't write transfer id (spooling message failed)",
			   gm_addr, error_to ? (error_to) : (NULL) , NULL);
			exit (1);
		}
	}
	else
	{
		fgets (str, sizeof (str) - 1, f);
		mid = atol (str) + 1;
		rewind (f);
	}
	fprintf (f, "%ld", mid);
	fclose (f);
	lock_delete(lockname.data);
	return mid;
}

static int 
handle_rmail(const char *inputfilename, char **adressen)
{
	/* 
	 * rmail empfnger1 empfnger2 ... 
	 * Anders gesagt: Wir haben keinen Absender, und nur die Defaultroute.
	 */
	int fd;
	int i;
	long id;
	const char *msgid=NULL;
	const char *from=NULL;
	const char *errors_to=NULL;

	/* 
	 * Nun die Nachricht lesen 
	 */
	if (inputfilename)
	{
		/* Dann Nachricht aus der Datei lesen */
		mbuf.len=0;
		if (rfc_read_message(inputfilename,&mbuf))
		{
			/* Hm? Eingabedatei nicht da? */
			struct stat st;
			log(LOG_ERR,"%s: %s\n", inputfilename, strerror (errno));
			/* don't bark because of missing files. testsuite does it,
			 * and manual tests do, too. */
			if (stat(inputfilename,&st)!=0 && 
				(errno==ENOENT || errno==ENAMETOOLONG))
				exit(1);
			errormail (NULL,
			   "weird: cannot read input file",
			   strerror (errno),
			   gm_addr, NULL , NULL);
			exit (1);
		}
	}
	else
	{
		/* Dann Nachricht von stdin lesen */
		mbuf.len=0;
		if (rfc_read_message_fd(0,&mbuf))
		{
			/* Hm? Eingabedatei nicht da? */
			log(LOG_ERR,_("stdin: %s\n"), strerror (errno));
			errormail (NULL,
			   "weird: cannot read stdin",
			   strerror (errno),
			   gm_addr, NULL , NULL);
			exit (1);
		}
	}
	from=rfc_find_header(mbuf.data,"from");
	if (!from)
		from=rfc_find_header(mbuf.data,"sender");
	if (!from)
		from=rfc_find_header(mbuf.data,"reply-to");
	if (from) {
		metalist_t *l=NULL;
		char *s=XSTRDUP(from);
		if (rfc_adress_to_metalist(0,&l,s) || !l) {
			/* Hmm???? Try to transport it, we can't bounce it! */
			log(LOG_ERR,"%s: can't convert %s\n", inputfilename, from);
			from=NULL;
		} else if (l->next) {
			log(LOG_ERR,"%s: throwing away excessive from addresses in %s\n", inputfilename ? inputfilename : "stdin", from);
			from=l->text;
		} else
			from=l->text;
	}
	errors_to=from;
	if (!from)
		from="[mroute: unknown sender]";

	if (!msgid)
   		msgid=rfc_find_header(mbuf.data,"message-id");
   	if (!msgid)
   		msgid="[mroute/rmail: unknown id]";

   	while (*adressen)
   	{
		uostr_t fname=UOSTR_INIT;
		i=0;
		do
		{
			char num[STR_ULONG];
			/* 
			 * eine Transfer-ID holen 
			 */
			id = get_transfer_id (errors_to);
			str_ulong(num,id);
			/* 
			 * In welches File spoolen? 
			 */
			uostr_xadd_cstrmulti(&fname,opts.spoolbase,"/",action,"/outgoing/mail",num,NULL);
			uostr_x0(&fname);
			fd=open(fname.data,O_WRONLY|O_CREAT|O_EXCL,0640);
		} while (i<100 && fd<0);

		if (fd<0)
		{
			/* Mail immer bouncen */
#define REASON "cannot open spool file, & out of memory"
			char *reason=concat("Cannot open spool file for mail to ",
				*adressen,NULL);
			bounce(from,reason ? reason : REASON , mbuf.data);
			msglog("# %s: %s, bounced",msgid,reason ? reason : REASON );
			errormail (NULL,
				reason ? reason : REASON,
			   strerror (errno),
			   gm_addr, NULL);
			adressen++;
			free(reason);
			uostr_freedata(&fname);
			continue;
#undef REASON
		}

		{
			char *p;
			/* uucp-Stil-Adressen "korrigieren". Die verursachen mehr Schaden 
			 * als Nutzen 
			 */
			if (!strchr(*adressen,'@') && (p=strchr(*adressen,'!'))!=NULL )
			{
				char *tmp=XMALLOC(strlen(*adressen)+1);
				*p++=0;
				sprintf(tmp,"%s@%s",p,*adressen);
				*adressen=tmp; /* slight memory leck */
			}
			if (!strchr(from,'@') && (p=strchr(from,'!'))!=NULL )
			{
				char *tmp=XMALLOC(strlen(from)+1);
				*p++=0;
				sprintf(tmp,"%s@%s",p,from);
				from=tmp; /* slight memory leck */
			}
		}
#define M1 "X-MausConnect-Env-To: "
#define M2 "\nX-MausConnect-Env-From: "
		if (xwrite(fd,M1,sizeof(M1)-1)<0 
			|| xwrite(fd,*adressen,strlen(*adressen))<0
			|| xwrite(fd,M2,sizeof(M2)-1)<0
			|| xwrite(fd,from,strlen(from))<0
			|| xwrite(fd,"\n",1)<0
			|| xwrite(fd,mbuf.data,mbuf.len)<0)
#undef M1
#undef M2
		{
			const char *reason2=
				"cannot write to spool file, & out of memory";
			char *reason=concat("Cannot write to spool file for mail to ",
				*adressen,NULL);
			close(fd);
			unlink(fname.data);
			/* Mail immer bouncen */
			bounce(from, reason ? reason : reason2, mbuf.data);
			errormail (NULL,
				reason ? reason : reason2,
			   strerror (errno),
			   gm_addr, NULL);
			msglog("# %s: %s, bounced",msgid, reason ? reason : reason2);
			free(reason);
		}
		else
			msglog("# %s: rmail spooled as %ld",msgid,id);
		close(fd);
		uostr_freedata(&fname);
		adressen++;
	}
	exit(0);
}

static int 
handle_mail(const char *inputfilename, const char *from, const char *to)
{
	int fd;
	int i;
	long id;
	uostr_t fname=UOSTR_INIT;
	const char *errors_to=NULL;
	const char *msgid=NULL;
	char *route;

	if (!to)
	{
		/* we should never be here? */
		log(LOG_ERR,_("receiver of mail not given"));
		exit(1);
	}
	if (!from) {
		log(LOG_ERR,_("sender of mail not given"));
		exit(1);
	}
	errors_to=XSTRDUP(from);
	if (opt_test)
	{
		route=route_it(from,to);
		if (route)
			printf(_("route for %s to %s via %s\n"),from, to,route);
		else
			printf(_("no route for %s to %s\n"),from,to);
		return 0;
	}

	/* 
	 * Nun die Nachricht lesen 
	 */
	if (inputfilename)
	{
		/* Dann Nachricht aus der Datei lesen */
		mbuf.len=0;
		if (rfc_read_message(inputfilename,&mbuf))
		{
			/* Hm? Eingabedatei nicht da? */
			struct stat st;
			log(LOG_ERR,"%s: %s\n", inputfilename, strerror (errno));
			/* don't bark because of missing files. testsuite does it,
			 * and manual tests do, too. */
			if (stat(inputfilename,&st)!=0 && 
				(errno==ENOENT || errno==ENAMETOOLONG))
				exit(1);
			errormail (NULL,
			   "weird: cannot read input file",
			   strerror (errno),
			   gm_addr, *errors_to ? (errors_to) : (NULL) , NULL);
			exit (1);
		}
	}
	else
	{
		/* Dann Nachricht von stdin lesen */
		mbuf.len=0;
		if (rfc_read_message_fd(0,&mbuf))
		{
			/* Hm? Eingabedatei nicht da? */
			log(LOG_ERR,"stdin: %s\n", strerror (errno));
			errormail (NULL,
			   "weird: cannot read stdin",
			   strerror (errno),
			   gm_addr, *errors_to ? (errors_to) : (NULL) , NULL);
			exit (1);
		}
	}

	if (!msgid)
   		msgid=rfc_find_header(mbuf.data,"message-id");
   	if (!msgid)
   		msgid="[mroute: unknown id]";

	/* 
	 * Die Nachricht ist jetzt also im Puffer. Nun routen.
	 */
	if (is_routing)
	{
		route=route_it(from,to);
		if (!route)
		{
			msglog("# %s: no route %s -> %s",
				msgid,from,to);
		}
		if (!route && opts.reroute_host)
		{
			/* Mail via eine andere Kiste schicken */
			/* XXX funktioniert nur wenn 'rmail' vorhanden ist */
			uostr_t cmd=UOSTR_INIT;
			char *p;
			FILE *wfd;
			uostr_xadd_cstr(&cmd,"rmail ");
			uostr_xadd_cstr(&cmd,to);
			p=memchr(cmd.data,0,cmd.len);
			if (p)
				*p='%';
			uostr_xadd_char(&cmd,'@');
			uostr_xadd_cstr(&cmd,opts.reroute_host);
			uostr_x0(&cmd);
			wfd=popen(cmd.data,"w");
			if (wfd)
			{
				fwrite(mbuf.data,strlen(mbuf.data),1,wfd);
				msglog("! %s: rerouted via %s",msgid,opts.reroute_host);
				exit(0);
			}
			msglog("# %s: rerouting failed with %s",msgid,strerror(errno));
		}
		if (!route)
		{
			struct bounce_info b;
			memset(&b,0,sizeof(b));
			b.orig_id=msgid;
			b.orig_subject=rfc_find_header(mbuf.data,"subject");
			b.orig_to=to;
			b.status_code="5.7.1"; /* Delivery not authorized, message refused */
				/* b.remote_mta=.. keiner beteiligt */
			b.diagnostic_code=concat("x-local; no route from ",
									 from," to ", to, NULL);
			XNULLCHECK(b.diagnostic_code);
			new_bounce(from,NULL,mbuf.data,&b);
			msglog("# %s: bounced",msgid);
			exit(0);
		}
		action=route;
	} else if (!is_allowed(from)) {
		struct bounce_info b;
		memset(&b,0,sizeof(b));
		b.orig_id=msgid;
		b.orig_subject=rfc_find_header(mbuf.data,"subject");
		b.orig_to=to;
		b.status_code="5.7.1"; /* Delivery not authorized, message refused */
		/* b.remote_mta=.. keiner beteiligt */
		b.diagnostic_code=concat("x-local; ", from,
								 " is not allowed to use the gateway", NULL);
		XNULLCHECK(b.diagnostic_code);
		new_bounce(from,NULL,mbuf.data,&b);
		msglog("# %s: bounced",msgid);
		exit(0);
	}

	i=0;
	do
	{
		char num[STR_ULONG];
		/* 
		 * eine Transfer-ID holen 
		 */
		id = get_transfer_id (errors_to);
		str_ulong(num,id);
		/* 
		 * In welches File spoolen? 
		 */
		uostr_xdup_cstrmulti(&fname,opts.spoolbase,"/",action,"/outgoing/mail",num,NULL);
		uostr_x0(&fname);
		fd=open(fname.data,O_WRONLY|O_CREAT|O_EXCL,0640);
	} while (i<100 && fd<0);


	if (fd<0)
	{
		/* Mail immer bouncen */
		bounce(from,"Cannot open spool file",
			mbuf.data);
		msglog("# %s: cannot open spoolfile, bounced",msgid);
		errormail (NULL,
		   "weird: cannot open spool file",
		   strerror (errno),
		   gm_addr, NULL);
		exit(0);
	}

	{
		char *p;
		/* uucp-Stil-Adressen "korrigieren". Die verursachen mehr Schaden 
		 * als Nutzen 
		 */
		if (!strchr(to,'@') && (p=strchr(to,'!'))!=NULL )
		{
			char *tmp=XMALLOC(strlen(to)+1);
			*p++=0;
			sprintf(tmp,"%s@%s",p,to);
			to=tmp; /* slight memory leck */
		}
		if (!strchr(from,'@') && (p=strchr(from,'!'))!=NULL )
		{
			char *tmp=XMALLOC(strlen(from)+1);
			*p++=0;
			sprintf(tmp,"%s@%s",p,from);
			from=tmp; /* slight memory leck */
		}
	}

#define M1 "X-MausConnect-Env-To: "
#define M2 "\nX-MausConnect-Env-From: "
	if (xwrite(fd,M1,sizeof(M1)-1)<0 
		|| xwrite(fd,to,strlen(to))<0
		|| xwrite(fd,M2,sizeof(M2)-1)<0
		|| xwrite(fd,from,strlen(from))<0
		|| xwrite(fd,"\n",1)<0
		|| xwrite(fd,mbuf.data,mbuf.len)<0)
#undef M1
#undef M2
	{
		const char *es=strerror(errno);
		close(fd);
		unlink(fname.data);
		/* Mail immer bouncen */
		bounce(from,"Cannot spool message", mbuf.data);
		errormail (NULL,
		   "weird: cannot spool mail",
		   es,
		   gm_addr, NULL);
		msglog("# %s: cannot spool messaged, bounced: %s",msgid,es);
		exit(0);
	}

	msglog("# %s: mail spooled as %ld",msgid,id);
	close(fd);
	exit(0);
}

static int 
handle_news(const char *inputfilename)
{
	char *from;
	metalist_t *ml;
	char *route;
	const char *msgid;
	long id;
	static uostr_t fname=UOSTR_INIT;
	int ok;
	int fd;
	int i;

	/* Bei News mssen alle wesentlichen Daten aus der Nachricht entnommen
	 * werden.
	 */
	if (inputfilename == NULL) {
		mbuf.len=0;
    	if (rfc_read_message_fd(0,&mbuf)) {
			/* Hm? Eingabedatei nicht da? */
			const char *es=strerror(errno);
			log(LOG_ERR,_("stdin: %s\n"), es);
			return (1);
    	}
    } else {
		/* Dann Nachricht aus der Datei lesen */
		mbuf.len=0;
    	if (rfc_read_message(inputfilename,&mbuf)) {
			/* Hm? Eingabedatei nicht da? */
			log(LOG_ERR,"%s: %s\n", inputfilename, strerror (errno));
			return 1;
    	}
    }
	msgid=rfc_find_header(mbuf.data,"message-id");
	if (!msgid) {
		log(LOG_ERR,"news without Message-Id-Header: %s\n", inputfilename ? inputfilename : "stdin");
		return 1;
	}
    /* 
     * Die Nachricht ist also vorhanden 
     * Jetzt Absender und Gruppe feststellen.
     */
    from=rfc_find_header(mbuf.data,"from");
    if (!from) {
		log(LOG_ERR,"news without From-Header: %s\n", inputfilename ? inputfilename : "stdin");
		return 1;
    } else {
		ml=NULL;
		if (rfc_adress_to_metalist(1,&ml,from)||!ml || !ml->text) {
			log(LOG_ERR,"cannot parse From-Header (%s) from %s\n", from,
				inputfilename ? inputfilename : "stdin");
			return 1;
		}
    	from=ml->text;
	}
    /* no good: some adresses can get longer, especially fido: strcpy(from,ml->text); */

	if (is_routing) {
		/* das ist _sicher_ ein "no match" fr alle User, darauf kann nur ein * passen */
		route=route_it(from,NULL);
		if (opt_test) {
			if (route)
				printf(_("route for news from %s via %s\n"),from, route);
			else
				printf(_("no route for news from %s\n"),from);
			return 0;
		}
	    if (!route) {
	    	/* das bedeutet, da wir noch nicht einmal wissen, ob
	    	 * die Nachricht ber ein Gate oder lokal kam. Die 
	    	 * Nachricht darf auch nicht "auf Verdacht" gebounced
	    	 * werden (wg. Gatewaybetrieb).
	    	 * Der Fall, da keine Route gefunden wird, kommt sehr hufig
	    	 * vor, wenn man Nachrichten aus dem Usenet bekommt mit
	    	 * Crosspostings in Maus-Gruppen. Von daher kann man die
	    	 * Warnung ausschalten :-)
	    	 */
	    	if (opts.warn_news_routing)
				log(LOG_WARNING,"cannot find a route for news %s from %s",msgid,from);
			msglog("# %s: no route, from %s",msgid,from);
	    	return 1;
	    }
	    action=route;
	} else if (!is_allowed(from)) {
		if (opts.warn_news_routing)
			log(LOG_WARNING,"%s is not allowed to send news %s through %s",from,msgid,action);
		msglog("# %s: no route, from %s",msgid,from);
		return 1;
	}

	/* Diverse Formen von Control-Headern bearbeiten */
	{	
		char *control;
		char *also_control=NULL;
		char *supersedes=NULL;
		char *replaces=NULL;
		int exit_after=0;
		control=rfc_find_header(mbuf.data,"control");
		also_control=rfc_find_header(mbuf.data,"also-control");
		supersedes=rfc_find_header(mbuf.data,"supersedes");
		replaces=rfc_find_header(mbuf.data,"replaces");
		if (control && !strncasecmp(control,"cancel ",7) && opts.honor_cancels) {
			work_cancel(control+7,from);
			exit_after=1;
		} else if (control)
			exit_after=1;
		if (supersedes && opts.honor_cancels)	{
			work_cancel(supersedes,from);
			msglog("  %s: supersedes added",msgid);
		}
		if (replaces && opts.honor_cancels)	{
			work_cancel(replaces,from);
			msglog("  %s: replaces added",msgid);
		}
		if (also_control && !strncasecmp(also_control,"cancel ",7) && opts.honor_cancels) {
			msglog("  %s: also-control cancel added",msgid);
			work_cancel(also_control+7,from);
		}
		if (exit_after)
			return 0;
	}

	ok=0;
	fd=-1; /* keep gcc happy */
	i=0;
	do
	{
		char num[STR_ULONG];
		/* 
		 * eine Transfer-ID holen 
		 */
		id = get_transfer_id (NULL);
		/* 
		 * In welches File spoolen? 
		 */
		str_ulong(num,id);
		uostr_xdup_cstrmulti(&fname,opts.spoolbase,"/",action,"/outgoing/",num,NULL);
		uostr_x0(&fname);
		if (!inputfilename)
		{
			fd=open(fname.data,O_WRONLY|O_CREAT|O_EXCL,0640);
			if (fd>=0)
				ok=1;
		}
		else
		{
#ifdef HAVE_SYMLINK
			if (symlink(inputfilename,fname.data)==0)
				ok=1;
			else
#endif
				if (file_copy(inputfilename,fname.data))
					ok=1;
		}
	} while (i<10 && !ok);

	if (!ok)
	{
		if (!inputfilename)
			log(LOG_ERR,"cannot spool news from stdin, open failed: %s", strerror(errno));
		else
			log(LOG_ERR,"cannot spool news from %s, symlink/file_copy failed: %s", 
				inputfilename, strerror(errno));
		msglog("# %s: cannot spool",msgid);
		return(1);
	}
	if (inputfilename)
	{
		msglog("# %s: news spooled as %ld",msgid,id);
		return 0;
	}
	if (xwrite(fd,mbuf.data,mbuf.len)<0)
	{
		log(LOG_ERR,"cannot spool news from stdin, write failed: %s", strerror(errno));
		close(fd);
		unlink(fname.data);
		msglog("# %s: cannot spool",msgid);
		return 1;
	} 
	close(fd);
	msglog("# %s: news spooled as %ld",msgid,id);
	return 0;
}

static int 
handle_rnews(const char *inputfilename)
{
	uostr_t fname=UOSTR_INIT;
	int fdin;
	int fdout;
	int i;
	unsigned long id;
	char *buf;
	size_t got;
	buf=malloc(256*1024);
	if (!buf)
		buf=XMALLOC(64*1024);

	/* einen Newsbatch nehmen und komplett spoolen */

	i=0;
	do
	{
		char num[STR_ULONG];
		/* 
		 * eine Transfer-ID holen 
		 */
		id = get_transfer_id (NULL);
		str_ulong(num,id);
		/* 
		 * In welches File spoolen? 
		 */
		uostr_xdup_cstrmulti(&fname,opts.spoolbase,"/",action,"/outgoing/nb",num,NULL);
		uostr_x0(&fname);
		fdout=open(fname.data,O_WRONLY|O_CREAT|O_EXCL,0640);
		i++;
	} while (i<10 && fdout<0);

	if (fdout<0)
	{
		const char *es=strerror(errno);
		errormail (NULL,
			"weird: cannot spool newsbatch",
			es,
		   gm_addr, NULL);
		msglog("# [%ld]: cannot open spoolfile: %s",id,es);
		return(1);
	}

	if (inputfilename == NULL)
		fdin=STDIN_FILENO;
	else
	{
		fdin=open(inputfilename,O_RDONLY);
		if (fdin<0)
		{
			const char *es=strerror(errno);
			struct stat st;
			/* Hm? Eingabedatei nicht da? */
			log(LOG_ERR,"%s: %s\n", inputfilename, es);
			msglog("# [%ld]: cannot open inputfile %s: %s",id,
				   inputfilename, es);
			unlink(fname.data);
			if (stat(inputfilename,&st)!=0
				&& (errno==ENOENT || errno==ENAMETOOLONG))
				return(1);
			errormail (NULL,
			   "weird: cannot read input file",
			   es,
			   gm_addr, NULL);
			return 1;
    	}
    }

    while ((got=read(fdin,buf,sizeof(buf)))>0)
    {
		if (xwrite(fdout,buf,got)<0)
		{
			const char *es=strerror(errno);
			close(fdout);
			unlink(fname.data);

			errormail (NULL,
			   "weird: cannot spool newsbatch",
			   es,
			   gm_addr, NULL);
			msglog("# [%ld]: cannot write to spoolfile: %s",id, es);
			return 1;
		}
    }
	msglog("# [%ld]: rnews spooled",id);
	close(fdin);
	return 0;
}

static void
test_routing(void)
{
	char from[1024];
	char to[1024];
	char *p;
	char *route;
	while (1)
	{
		printf(_("from> "));
		if (!fgets(from,sizeof(from),stdin))
			break;
		p=strpbrk(from,"\r\n");
		if (p)
			*p='\0';
		if (!*from)
			break;
		printf(_("  to> "));
		if (!fgets(to,sizeof(to),stdin))
			break;
		p=strpbrk(to,"\r\n");
		if (p)
			*p='\0';
		if (!*to)
			break;
		route=route_it(strtrim(from),strtrim(to));
		if (route)
			printf(_(" via %s\n"),route);
		else
			printf(_(" no way\n"));
	}
}

#define WHAT_MAIL 1
#define WHAT_RMAIL 2
#define WHAT_RSMTP 3
#define WHAT_NEWS 4
#define WHAT_RNEWS 5
#define WHAT_ROUTE_SHOW 6
#define WHAT_ROUTE_TEST 8


#define MAX_SETVARS 10

static const char *opt_cnffname;
static const char *opt_to=NULL;
static const char *opt_from=NULL;
static int what_i_do=0;
static const char *opt_inputfilename=NULL;


static uogetopt_t myopts[]={
	{'a',"action",UOGO_STRING|UOGO_HIDDEN,&action,0,
	"Channel to use (old name)","CHANNEL"},
	{'C',"channel",UOGO_STRING,&action,0,
	"Channel to use","CHANNEL"},
	{'c',"cnffile",UOGO_STRING,&opt_cnffname,0,
	"Configuration file to use","FILE"},
	{'c',"cnf-file",UOGO_STRING|UOGO_HIDDEN,&opt_cnffname,0,
	"Configuration file to use","FILE"},
	{'f',"from",UOGO_STRING,&opt_from,0,
	"sender address\n"
	"Use this to specify the sender address\n"
	"Only used in --mail mode","ADDRESS"},
	{'i',"input",UOGO_STRING,&opt_inputfilename,0,
	"File to read input from","FILE"},
	{'m',"mail",UOGO_FLAG,&what_i_do,WHAT_MAIL,
	"spool a single mail",NULL},
	{  0,"rmail",UOGO_FLAG,&what_i_do,WHAT_RMAIL,
	"work as rmail",NULL},
	{  0,"rnews",UOGO_FLAG,&what_i_do,WHAT_RNEWS,
	"work as rnews\n"
	"Receives a compressed or uncompressed\n"
	"batch of news",NULL},
	{  0,"news",UOGO_FLAG,&what_i_do,WHAT_NEWS,
	"receive a single news",NULL},
	{  0,"rsmtp",UOGO_FLAG,&what_i_do,WHAT_RSMTP,
	"work as rsmtp\n"
	"Receives a rsmtp mail batch",NULL},
	{'s',"set",UOGO_CALLBACK,&setvarfn,0,
	"override variables in the configuration file","VAR=VALUE"},
	{'T',"to",UOGO_STRING,&opt_to,0,
	"recipient address\n"
	"Use this to specify the recipient address\n"
	"Only uses in --mail mode","ADDRESS"},

	{  0,"route-show",UOGO_FLAG,&what_i_do,WHAT_ROUTE_SHOW,
	"show routes", NULL},
	{  0,"route-test",UOGO_FLAG,&what_i_do,WHAT_ROUTE_TEST,
	"test routes", NULL},
	{  0,"test",UOGO_FLAG,&opt_test,1,
	"enter test mode, do not deliver.\n"
	"Only works in --mail and --news mode.", NULL},

	{ 0 ,"copyright", UOGO_PRINT|UOGO_HIDDEN,0,0,
	"Copyright (C) 1999 Uwe Ohse\n"
	PACKAGE " comes with ABSOLUTELY NO WARRANTY,\n"
	"to the extent permitted by law.\n"
	"You may redistribute copies of " PACKAGE "\n"
	"under the terms of the GNU General Public License.\n"
	"For more information about these matters, see the file named COPYING.\n"
	"You may redistribute copies of " PACKAGE "\n",0},
									
	{0,0,0,0,0,0,0} /* --help and --version */
};

int
main (int argc, char **argv)
{
	int ret;
   	BOOLEAN override_action=FALSE;

	const char *slash;

	program_name=argv[0];
	slash=strrchr(program_name,'/');
	if (slash)
		program_name=slash+1;

	uogetopt (program_name, PACKAGE, VERSION, &argc, argv, uogetopt_out,
	"usage: mroute [options] [channelname]\n"
    "Spool messages for transport to MausTausch-BBS\n",
		myopts,
	"\nReport bugs to <uwe-mauscon@bulkmail.ohse.de>");

	if (opt_cnffname) {
		if (geteuid()!=getuid() && getuid()!=0) {
			const char *er=": --cnffile option not allowed for you\n";
			log (LOG_ERR, _("not for you (possible try to hack?)"));
			write(2,program_name,strlen(program_name));
			write(2,er,strlen(er));
			exit(2);
		}
	} else	
		opt_cnffname=SYSCONFDIR "/" CNFNAME;
	if (setvars) {		
		if (geteuid()!=getuid() && getuid()!=0) {
			const char *er=": --set option not allowed for you\n";
			log (LOG_ERR, _("not for you (possible try to hack?)"));
			write(2,program_name,strlen(program_name));
			write(2,er,strlen(er));
			exit(2);
		}
	}

	if (what_i_do==0)
	{
		if (!strcmp(program_name,"rnews") || !strcmp(program_name,"mrnews"))
			what_i_do=WHAT_RNEWS;
		if (!strcmp(program_name,"mmail"))
			what_i_do=WHAT_MAIL;
		if (!strcmp(program_name,"rmail") || !strcmp(program_name,"mrmail"))
			what_i_do=WHAT_RMAIL;
		if (!strcmp(program_name,"rsmtp") || !strcmp(program_name,"mrsmtp"))
			what_i_do=WHAT_RSMTP;
		if (!strcmp(program_name,"mnews"))
			what_i_do=WHAT_NEWS;

		if (!what_i_do)
		{
			log(LOG_ERR, _("%s: don't know what to do"),program_name);
			exit(1);
		}
	}

	switch(what_i_do)
	{
	case WHAT_RMAIL:
	 	/* rmail to1 [to2 ...] */
		if (!argv[1])
		{
			log (LOG_ERR, _("no address given"));
			exit(2);
		}
		if (!action)
		{
			action="rmail";
   			override_action=TRUE;
   		}
		break;
	case WHAT_RSMTP:
	 	/* rsmtp [action] */
	 	if (argv[1])
	 	{
			if (argv[2] || action)
			{
				log (LOG_ERR, _("too many arguments"));
				exit (2);
			}
    		action=argv[1];
	 	}
    	else if (!action)
    	{
    		action="rsmtp"; /* Default */
   			override_action=TRUE;
   		}
		break;
	case WHAT_RNEWS:
		/* rnews [inputfile] */
		if (!action)
		{
			action="rnews";
			override_action=TRUE;
		}
		if (argv[1])
		{
			if (argv[2])
			{
				log (LOG_ERR, _("too many arguments"));
				exit (2);
			}
    		if (opt_inputfilename)
				log (LOG_WARNING, 
					 _("warning: given filename overrides --input"));
			opt_inputfilename=argv[1];
			if (!strcmp(opt_inputfilename,"-"))
				opt_inputfilename=NULL;
    	}
    	break;
	case WHAT_MAIL:
	case WHAT_NEWS:
		/* mail {action | "route"} */
	 	/* news {action | "route"} */
		if (!argv[1])
		{
			if (!action) {
				log (LOG_ERR, _("channel not given"));
				exit(2);
			}
		}
		else if (argv[2] || action)
		{
			log (LOG_ERR, _("too many arguments"));
			exit (2);
		}
		else 
			action=argv[1];
		break;
	case WHAT_ROUTE_SHOW:
		if (!argv[1]) {
			if (!action) {
				log (LOG_ERR, _("channel not given"));
				exit(2);
			}
		} else if (argv[2] || action) {
			log (LOG_ERR, _("too many arguments"));
			exit (2);
		}
		action=argv[1];
		break;
	case WHAT_ROUTE_TEST:
		action="route";
		break;
	}

	if (!strcasecmp(action,"route"))
		is_routing=1;

    if (do_conf(opt_cnffname,action ? action : "route", 
    	setvarscount ? setvars : NULL,setvarscount ))
    	exit(1);

	if (override_action)
	{
		if (what_i_do== WHAT_RMAIL)
			action=opts.rmail_route;
		else if (what_i_do==WHAT_RSMTP)
			action=opts.rsmtp_route;
		else if (what_i_do==WHAT_RNEWS)
			action=opts.rnews_route;
		if (!action)
		{
			log (LOG_ERR, _("indirect channel not set"));
			exit(2);
		}
	}

	gm_addr=xconcat("=",opts.gatemaster,NULL);

	openlog (program_name, 0, opts.sl_facility);
	setlogmask(LOG_UPTO(opts.sl_priority));
	open_msglog();

	if (route_open(opts.spoolbase))
	{
		log(LOG_ERR,_("cannot open routing database"));
		exit(1);
	}

	if (what_i_do==WHAT_ROUTE_SHOW)
	{
		ret=route_show(action);
	}
	else if (what_i_do==WHAT_ROUTE_TEST)
	{
		test_routing();
		ret=0;
	}
	else if (what_i_do==WHAT_RMAIL)
	{
		ret=handle_rmail(opt_inputfilename,&argv[1]);
	}
	else if (what_i_do==WHAT_MAIL)
	{
		ret=handle_mail(opt_inputfilename,opt_from,opt_to);
	}
	else if (what_i_do==WHAT_RSMTP)
	{
		ret=handle_rsmtp(opt_inputfilename);
	}
	else if (what_i_do==WHAT_RNEWS)
		ret=handle_rnews(opt_inputfilename);	
	else /* WHAT_NEWS */
		ret=handle_news(opt_inputfilename);	
	if (got_cancels)
		cancel_end(0,1);

	closelog ();
	close_msglog();

	exit (ret);
}

static void 
cleanup_empfaenger(struct empfaenger **anker)
{
	struct empfaenger *p,*next;
	p=*anker;
	while (p)
	{
		next=p->next;
		if (p->adr)
			free(p->adr);
		free(p);
		p=next;
	}
	*anker=NULL;
}

static int 
handle_rsmtp(const char *inputfilename)
{
	FILE *f=stdin;
	char *p;
	char buf[1024];
	size_t l;
	int retcode=0;
	BOOLEAN got_greeting=FALSE;
	BOOLEAN got_ehlo=FALSE;
	BOOLEAN got_data=FALSE;
	unsigned long errors=0;
	char from[1024]="";
	BOOLEAN got_from=FALSE;
	struct empfaenger *anker=NULL;
	uostr_t body=UOSTR_INIT;
	BOOLEAN in_header=FALSE;

	if (inputfilename)
	{
		f=fopen(inputfilename,"r");
		if (!f)
		{
			log(LOG_ERR,_("%s: cannot open %s: %s"), "mrsmtp",
				inputfilename,strerror(errno));
			return -1;
		}
	}

	while (1)
	{
		if (!fgets(buf,sizeof(buf),f))
		{
			retcode=-1;
			log(LOG_ERR,_("mrsmtp: unexpected end of transmission: %s"),	
				strerror(errno));
			break;
		}
		p=strpbrk(buf,"\r\n");
		if (p)
			*p='\0';
puts(buf);
		if (!got_greeting)
		{
			if (!strncasecmp(buf,"HELO ",5))
			{
				got_greeting=TRUE;
			}
			else if (!strncasecmp(buf,"EHLO ",5))
			{
				got_greeting=TRUE;
				got_ehlo=TRUE;
			}
			else 
				errors++; 
			continue;
		}
		if (!got_data)
		{
			if (!strncasecmp(buf,"MAIL FROM:",10))
			{
				char *adr=buf+10;
				while (isspace(*adr))
					adr++;
				if (*adr=='<')
				{
					adr++;
					while (isspace(*adr))
						adr++;
				}
				p=strpbrk(adr,"\t ");
				if (p)
					*p=0;
				strcpy(from,adr);
				p=strchr(from,'>');
				if (p)
					*p=0;
				cleanup_empfaenger(&anker);
				got_from=TRUE;
				continue;
			}
			if (!strncasecmp(buf,"RCPT TO:",8))
			{
				char *adr=buf+8;
				struct empfaenger *new;

				if (!got_from)
				{
					/* nicht nur ein kleines Problem: Es gibt noch nicht 
					 * mal eine Adresse, an die die Fehlermeldung geschickt
					 * werden kann.
					 */
					log(LOG_NOTICE,_("got smtp without MAIL FROM"));
					/* 
					 * Bser Hack, mglicherweise ein Fehler?
					 */
					strcpy(from,"unknown");
					continue;
				}
				while (isspace(*adr))
					adr++;
				if (*adr=='<')
				{
					adr++; /* wegen '<' */
					while (isspace(*adr))
						adr++;
				}
				p=strpbrk(adr,"\t ");
				if (p)
					*p=0;
				p=strchr(adr,'>');
				if (p)
					*p=0;
				new=XMALLOC(sizeof(struct empfaenger));
				new->adr=XSTRDUP(adr);
				new->next=anker;
				anker=new;
				continue;
			}
			if (!strcasecmp(buf,"DATA"))
			{
				if (!anker)
				{
					log(LOG_NOTICE, _("no receiver given -> skipped"));
					continue;
				}
				if (!got_from)
				{
					log(LOG_NOTICE,_("no sender given -> skipped"));
					continue;
				}
				in_header=TRUE;
				got_data=1;
				continue;
			}
			if (!strcasecmp(buf,"QUIT"))
			{
				cleanup_empfaenger(&anker);
				break;
			}
			continue;
		}
		/* got_data gesetzt */
		if (*buf=='.')
		{
			if (!buf[1])
			{
				uostr_x0(&body);
				finish_rsmtp(from,anker,&body);
				got_data=FALSE;
				body.len=0;
				body.data[0]=0;
				continue;
			}
			p=buf+1;
		}
		else
			p=buf;
		l=strlen(p);
		if (!*p)
			in_header=FALSE;
		if (in_header && isspace(*p) && body.len>2)
			/* Fortsetzungszeile */
			uostr_cut(&body,-2);
		uostr_xadd_cstr(&body,p);
		uostr_xadd_mem(&body,"\r\n",2);
	}
	uostr_freedata(&body);
	return retcode;
}


static void 
finish_rsmtp(char *from, struct empfaenger *emp, uostr_t *message)
{
	char env_buf[2048];
	char fname[PATH_MAX];
	long id;
	const char *msgid;
	struct empfaenger *err_route_list=NULL;
	struct empfaenger *err_spool_list=NULL;
	char *route;

	msgid=rfc_find_header(message->data,"message-id");
   	if (!msgid)
   		msgid="[mroute: unknown id]";

	while (emp)
	{
		route=route_it(from,emp->adr);
		if (!route && opts.reroute_host)
		{
			/* Mail via eine andere Kiste schicken */
			char adr[1024];
			char cmd[2048];
			char *p;
			FILE *wfd;
			strcpy(adr,emp->adr);
			p=strchr(adr,'@');
			if (p)
				*p='%';
			strcat(adr,"@");
			strcat(adr,opts.reroute_host);
			sprintf(cmd,"rmail '%s'",adr);
			wfd=popen(cmd,"w");
			if (wfd)
			{
				fwrite(message->data,strlen(message->data),1,wfd);
				msglog("# %s: routing failed, rerouted via %s",
					msgid,opts.reroute_host);
				pclose(wfd);
			}
			else
				msglog("# %s: routing failed, rerouting also with %s",
					msgid,strerror(errno));
		}
		else if (!route)
		{
			struct empfaenger *n=XMALLOC(sizeof(struct empfaenger));
			n->adr=emp->adr;
			n->next=err_route_list;
			err_route_list=n;
			msglog("# %s: routing failed", msgid);
		}
		else 
		{
			int i=0;
			int fd;
			action=route;
			do
			{
				/* 
				 * eine Transfer-ID holen 
				 */
				id = get_transfer_id (NULL);
				/* 
				 * In welches File spoolen? 
				 */
				sprintf (fname, "%s/%s/outgoing/mail%ld", 
					opts.spoolbase, action, (long) id);
				fd=open(fname,O_WRONLY|O_CREAT|O_EXCL,0640);
			} while (i<100 && fd<0);

			if (fd<0)
			{
				if (from && *from)
				{
					struct empfaenger *n=XMALLOC(sizeof(struct empfaenger));
					n->adr=emp->adr;
					n->next=err_spool_list;
					err_spool_list=n;
					errormail (NULL,
					   "weird: cannot open spool file",
					   strerror (errno),
					   gm_addr, NULL);
				}
				msglog("# %s: cannot spool as %ld",msgid,id);
			}
			else
			{

				sprintf(env_buf,
					"X-MausConnect-Env-To: %s\n"
					"X-MausConnect-Env-From: %s\n",
					emp->adr,
					from);

				if (xwrite(fd,env_buf,strlen(env_buf))<0 
					|| xwrite(fd,message->data,message->len)<0)
				{
					struct empfaenger *n=XMALLOC(sizeof(struct empfaenger));
					n->adr=emp->adr;
					n->next=err_spool_list;
					err_spool_list=n;
					unlink(fname);
					errormail (NULL,
					   "weird: cannot spool mail",
					   strerror (errno),
					   gm_addr, NULL);
					msglog("# %s: cannot spool as %ld (write error)",msgid,id);
				}
				close(fd);
				msglog("  %s: rsmtp spooled as %ld",msgid,id);
			}

		}
		emp=emp->next;
	}

	if (err_route_list || err_spool_list)
	{
		struct empfaenger *next;
		char buf[2048];
		size_t bufl=0;
		size_t l;

		if (err_route_list)
		{
			bufl+=sprintf(buf,"You are not allowed to use the gateway to:");
			while (err_route_list)
			{
				l=strlen(err_route_list->adr);
				if (l+3+bufl<sizeof(buf))
					bufl+=sprintf(buf+bufl,"%s\n",err_route_list->adr);
				next=err_route_list;
				free(err_route_list->adr);
				free(err_route_list);
				err_route_list=next;
			}
		}
		if (err_spool_list)
		{
			bufl+=sprintf(buf,"Couldn't spool message to:");
			while (err_spool_list)
			{
				l=strlen(err_spool_list->adr);
				if (l+3+bufl<sizeof(buf))
					bufl+=sprintf(buf+bufl,"%s\n",err_spool_list->adr);
				next=err_spool_list;
				free(err_spool_list->adr);
				free(err_spool_list);
				err_spool_list=next;
			}
		}
		errormail (NULL,
		   "rsmtp prozessing error(s)",
		   buf,
		   gm_addr,from, NULL);
	}
}

