/* MausConnect: Nachrichtenformatskonverter fuer das MausNet
   Copyright (C) 1992 1993 1994 1995 Michael Kraehe, Uwe Ohse

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   srfcnews.c - send news in rfc-format 
*/

#include "config.h"
#include <sys/types.h>
#ifdef HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <conflib.h>
#include <limits.h>
#include <syslog.h>
#include <time.h>

#include "mausconlib.h"
#include "mauscon.h"
#include "metamsg.h"
#include "rfc.h"
#ifdef WANT_NNTP
#include "nntp.h"
#endif
#include "srfcnews.h"
#include "msg_io.h"
#include "uostr.h"


#ifdef WANT_NNTP
static BOOLEAN nntp_tried;
static mio_t *nntp_channel;
#endif

static mio_t *batch_channel;

BOOLEAN 
srfcnews(int only_convert, metamsg_t *msg)
{
	mio_retcode_t ret;
	if (!msg)
	{
		log(LOG_WARNING,_("srfcnews: no message"));
		/* Wenn man nichts versendet, kann man auch keine Fehler machen */
		return TRUE;
	}
	if (!msg->newsgroups || !msg->newsgroups->text)
	{
		log(LOG_ERR, _("srfcnews: no newsgroup given"));
		return FALSE;
	}
	if (only_convert)
	{
		rfc_print_message(msg,stdout,"\n",0);
		return TRUE;
	}
#ifdef WANT_NNTP
	if (!nntp_channel && !nntp_tried)
	{
		if (!dont_use_nntp)
		{
			uostr_t errbuf=UOSTR_INIT;
			if (MIO_OK!=mio_nntp.open(opts.newshost,"w",&nntp_channel,&errbuf))
			{
				log(LOG_ERR,_("cannot open nntp(%s): %s"),opts.newshost,errbuf.data);
				msglog("  <generic> cannot open nntp(%s): %s",
					opts.newshost,errbuf.data);
			}
			uostr_freedata(&errbuf);
		}
		nntp_tried=1;
	}
	if (nntp_channel)
	{
		ret=nntp_channel->fn->export(nntp_channel,msg);
		if (ret==MIO_OK)
			return TRUE;
		else if (ret==MIO_ERROR) {
			msglog("# %s: closing NNTP channel",msg->message_id);
			/* f_nntp meldet nur ERROR in den Fllen, in denen 
			 * weitermachen keinen Sinn hat.
			 */
			nntp_channel->fn->close(nntp_channel);
			nntp_channel=NULL;
		}
	}
#endif
	if (!batch_channel)
	{
		uostr_t fname=UOSTR_INIT;
		struct stat st;
		size_t pos;
		unsigned long cnt=(unsigned long)time(NULL);
		uostr_t errbuf=UOSTR_INIT;
		uostr_xadd_cstr(&fname,opts.spoolbase);
		uostr_xadd_char(&fname,'/');
		uostr_xadd_cstr(&fname,action);
		uostr_xadd_cstr(&fname,"newsbatch.");
		pos=fname.len;
		do
		{
			char numbuf[STR_ULONG];
			uostr_cut(&fname,pos);
			uostr_xadd_mem(&fname,numbuf,str_ulong(numbuf,cnt));
			uostr_0(&fname);
			cnt++;
		} while (stat(fname.data,&st)!=-1);	
		ret=mio_rnews.open(fname.data,"w",&batch_channel,&errbuf);
		uostr_freedata(&fname);
		if (ret!=MIO_OK)
		{
			log(LOG_ERR,_("cannot open news batchfile: %s"),errbuf.data);
			msglog("  <generic> cannot open batchfile: %s",errbuf.data);
			uostr_freedata(&errbuf);
			return FALSE;
		}
		uostr_freedata(&errbuf);
	}
	ret=batch_channel->fn->export(batch_channel,msg);
	if (ret==MIO_OK)
	{
		msglog("  %s: newsbatched",msg->message_id);
		return TRUE;
	}
	return FALSE;
}

void 
srfcnews_end_all(void)
{
#ifdef WANT_NNTP
	if (nntp_channel)
	{
		nntp_channel->fn->close(nntp_channel);
		nntp_channel=NULL;
	}
#endif
	if (batch_channel)
	{
		batch_channel->fn->close(batch_channel);
		batch_channel=NULL;
	}
}
