/*
 * Copyright (C) 1999 Uwe Ohse
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * As a special exception this source may be used as part of the
 * SRS project by CORE/Computer Service Langenbach
 * regardless of the copyright they choose.
 * 
 * Contact: uwe@ohse.de
 */
#include <stdlib.h>
#include <fcntl.h>
#include <sys/file.h>
#include <unistd.h>
#include <errno.h>
#include "uostr.h"
#include "uoio.h"
#include "uoconfig.h"

/* to 100% not in the spirit of conflib: see, i've learned a lesson! */
int 
uoconf_cstr(uoconfstr_t *u,const char *s)
{
	const char *p;
	size_t l;
	if (u->flag & UOCONF_FLAG_DONE) return 0;
	if (!s) return 0; 
	p=s;while(*p)p++;
	l=p-s+1; /* we do want the \0, too */
	if (!uostr_dup_mem(&(u->str),s,l)) return -1;
	u->flag|=UOCONF_FLAG_DONE;
	return 0;
}

int 
uoconf_environ(uoconfstr_t *u, const char *s)
{
	return uoconf_cstr(u,getenv(s));
}

int 
uoconf_firstline(uoconfstr_t *u,const char *fn, int strip, int commentchar)
{
	int fd;
	uoio_t i;
	if (u->flag & UOCONF_FLAG_DONE) return 0;
	fd=open(fn,O_RDONLY);
	if (fd==-1) {
		if (errno==ENOENT) return 0;
		return -1;
	}
	uoio_assign_r(&i,fd,read,0);
	while (1) {
		char *s,*p,*q;
		ssize_t len;
		int stripstart=0;
		int ok=1;
		if (strip) stripstart=1;
		len=uoio_getdelim_zc(&i, &s, '\n');
		if (len==-1||len==0) { int e=errno; uoio_destroy(&i); close(fd); errno=e; return len; }
		if (s[len-1]=='\n') {s[len-1]=0; len--;}
		/* process line */
		for (p=s,q=p+len;p!=q;) {
			/* no \0 allowed. could escape it, but why bother? And why should the upper layer
			 * want to unescape? */
			if (*p==0) { uoio_destroy(&i); close(fd); errno=ERANGE; return -1; }
			if (stripstart && (*p==' ' || *p=='\t')) {
				s++; p++; continue;
			}
			if (stripstart) stripstart=0;
			if (*p==commentchar) { break; }
			p++;
		}
		if (*p==commentchar && p==s) continue; /* comment char at start of line */ 
		if (strip) {
			while (p>s && (p[-1]=='\t' || p[-1]==' ')) p--;
		}
		/* s is start of value. p is at the character position behind the end or on s (in case of an empty line) */
		/* and the string need not be 0 terminated */
		if (!uostr_need(&(u->str),((size_t)(p-s+1)))) {
			uoio_destroy(&i);
			close(fd);
			return -1;
		}
		if (p>s && !uostr_dup_mem(&u->str,s,p-s)) ok=0;
		if (!uostr_add_mem(&u->str,"",1)) ok=0;
		uoio_destroy(&i);
		close(fd);
		if (!ok) 
			return -1;
		u->flag|=UOCONF_FLAG_DONE;
		return 0;
	}
}

